# Copyright (c) Xuechen Li. All Rights Reserved.
# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
This file is adapted from the privacy accounting procedure in Opacus', which in turn is adapted from tf-privacy.
Below is the original documentation in Opacus.

*Based on Google's TF Privacy:* https://github.com/tensorflow/privacy/blob/master/tensorflow_privacy/privacy/analysis
/rdp_accountant.py.
*Here, we update this code to Python 3, and optimize dependencies.*

Functionality for computing Renyi Differential Privacy (RDP) of an additive
Sampled Gaussian Mechanism (SGM).

Example:
    Suppose that we have run an SGM applied to a function with L2-sensitivity of 1.

    Its parameters are given as a list of tuples
    ``[(q_1, sigma_1, steps_1), ..., (q_k, sigma_k, steps_k)],``
    and we wish to compute epsilon for a given target delta.

    The example code would be:

    >>> max_order = 32
    >>> orders = range(2, max_order + 1)
    >>> rdp = np.zeros_like(orders, dtype=float)
    >>> for q, sigma, steps in parameters:
    >>>     rdp += privacy_analysis.compute_rdp(q, sigma, steps, orders)
    >>> epsilon, opt_order = privacy_analysis.get_privacy_spent(orders, rdp, delta)
"""

import math
import numpy as np
from scipy import special
from typing import List, Sequence, Union


########################
# LOG-SPACE ARITHMETIC #
########################


def _log_add(logx: float, logy: float) -> float:
    r"""Adds two numbers in the log space.

    Args:
        logx: First term in log space.
        logy: Second term in log space.

    Returns:
        Sum of numbers in log space.
    """
    a, b = min(logx, logy), max(logx, logy)
    if a == -np.inf:  # adding 0
        return b
    # Use exp(a) + exp(b) = (exp(a - b) + 1) * exp(b)
    return math.log1p(math.exp(a - b)) + b  # log1p(x) = log(x + 1)


def _log_sub(logx: float, logy: float) -> float:
    r"""Subtracts two numbers in the log space.

    Args:
        logx: First term in log space. Expected to be greater than the second term.
        logy: First term in log space. Expected to be less than the first term.

    Returns:
        Difference of numbers in log space.

    Raises:
        ValueError
            If the result is negative.
    """
    if logx < logy:
        raise ValueError("The result of subtraction must be non-negative.")
    if logy == -np.inf:  # subtracting 0
        return logx
    if logx == logy:
        return -np.inf  # 0 is represented as -np.inf in the log space.

    try:
        # Use exp(x) - exp(y) = (exp(x - y) - 1) * exp(y).
        return math.log(math.expm1(logx - logy)) + logy  # expm1(x) = exp(x) - 1
    except OverflowError:
        return logx


def _compute_log_a_for_int_alpha(q: float, sigma: float, alpha: int) -> float:
    r"""Computes :math:`log(A_\alpha)` for integer ``alpha``.

    Notes:
        Note that
        :math:`A_\alpha` is real valued function of ``alpha`` and ``q``,
        and that 0 < ``q`` < 1.

        Refer to Section 3.3 of https://arxiv.org/pdf/1908.10530.pdf for details.

    Args:
        q: Sampling rate of SGM.
        sigma: The standard deviation of the additive Gaussian noise.
        alpha: The order at which RDP is computed.

    Returns:
        :math:`log(A_\alpha)` as defined in Section 3.3 of
        https://arxiv.org/pdf/1908.10530.pdf.
    """

    # Initialize with 0 in the log space.
    log_a = -np.inf

    for i in range(alpha + 1):
        log_coef_i = (
            math.log(special.binom(alpha, i))
            + i * math.log(q)
            + (alpha - i) * math.log(1 - q)
        )

        s = log_coef_i + (i * i - i) / (2 * (sigma ** 2))
        log_a = _log_add(log_a, s)

    return float(log_a)


def _compute_log_a_for_frac_alpha(q: float, sigma: float, alpha: float) -> float:
    r"""Computes :math:`log(A_\alpha)` for fractional ``alpha``.

    Notes:
        Note that
        :math:`A_\alpha` is real valued function of ``alpha`` and ``q``,
        and that 0 < ``q`` < 1.

        Refer to Section 3.3 of https://arxiv.org/pdf/1908.10530.pdf for details.

    Args:
        q: Sampling rate of SGM.
        sigma: The standard deviation of the additive Gaussian noise.
        alpha: The order at which RDP is computed.

    Returns:
        :math:`log(A_\alpha)` as defined in Section 3.3 of
        https://arxiv.org/pdf/1908.10530.pdf.
    """
    # The two parts of A_alpha, integrals over (-inf,z0] and [z0, +inf), are
    # initialized to 0 in the log space:
    log_a0, log_a1 = -np.inf, -np.inf
    i = 0

    z0 = sigma ** 2 * math.log(1 / q - 1) + 0.5

    while True:  # do ... until loop
        coef = special.binom(alpha, i)
        log_coef = math.log(abs(coef))
        j = alpha - i

        log_t0 = log_coef + i * math.log(q) + j * math.log(1 - q)
        log_t1 = log_coef + j * math.log(q) + i * math.log(1 - q)

        log_e0 = math.log(0.5) + _log_erfc((i - z0) / (math.sqrt(2) * sigma))
        log_e1 = math.log(0.5) + _log_erfc((z0 - j) / (math.sqrt(2) * sigma))

        log_s0 = log_t0 + (i * i - i) / (2 * (sigma ** 2)) + log_e0
        log_s1 = log_t1 + (j * j - j) / (2 * (sigma ** 2)) + log_e1

        if coef > 0:
            log_a0 = _log_add(log_a0, log_s0)
            log_a1 = _log_add(log_a1, log_s1)
        else:
            log_a0 = _log_sub(log_a0, log_s0)
            log_a1 = _log_sub(log_a1, log_s1)

        i += 1
        if max(log_s0, log_s1) < -30:
            break

    return _log_add(log_a0, log_a1)


def _compute_log_a(q: float, sigma: float, alpha: float) -> float:
    r"""Computes :math:`log(A_\alpha)` for any positive finite ``alpha``.

    Notes:
        Note that
        :math:`A_\alpha` is real valued function of ``alpha`` and ``q``,
        and that 0 < ``q`` < 1.

        Refer to Section 3.3 of https://arxiv.org/pdf/1908.10530.pdf
        for details.

    Args:
        q: Sampling rate of SGM.
        sigma: The standard deviation of the additive Gaussian noise.
        alpha: The order at which RDP is computed.

    Returns:
        :math:`log(A_\alpha)` as defined in the paper mentioned above.
    """
    if float(alpha).is_integer():
        return _compute_log_a_for_int_alpha(q, sigma, int(alpha))
    else:
        return _compute_log_a_for_frac_alpha(q, sigma, alpha)


def _log_erfc(x: float) -> float:
    r"""Computes :math:`log(erfc(x))` with high accuracy for large ``x``.

    Helper function used in computation of :math:`log(A_\alpha)`
    for a fractional alpha.

    Args:
        x: The input to the function

    Returns:
        :math:`log(erfc(x))`
    """
    return math.log(2) + special.log_ndtr(-x * 2 ** 0.5)


def _compute_rdp(q: float, sigma: float, alpha: float) -> float:
    r"""Computes RDP of the Sampled Gaussian Mechanism at order ``alpha``.

    Args:
        q: Sampling rate of SGM.
        sigma: The standard deviation of the additive Gaussian noise.
        alpha: The order at which RDP is computed.

    Returns:
        RDP at order ``alpha``; can be np.inf.
    """
    if q == 0:
        return 0

    # no privacy
    if sigma == 0:
        return np.inf

    if q == 1.0:
        return alpha / (2 * sigma ** 2)

    if np.isinf(alpha):
        return np.inf

    return _compute_log_a(q, sigma, alpha) / (alpha - 1)


def compute_rdp(
    q: float, noise_multiplier: float, steps: int, orders: Union[Sequence[float], float]
) -> Union[List[float], float]:
    r"""Computes Renyi Differential Privacy (RDP) guarantees of the
    Sampled Gaussian Mechanism (SGM) iterated ``steps`` times.

    Args:
        q: Sampling rate of SGM.
        noise_multiplier: The ratio of the standard deviation of the
            additive Gaussian noise to the L2-sensitivity of the function
            to which it is added. Note that this is same as the standard
            deviation of the additive Gaussian noise when the L2-sensitivity
            of the function is 1.
        steps: The number of iterations of the mechanism.
        orders: An array (or a scalar) of RDP orders.

    Returns:
        The RDP guarantees at all orders; can be ``np.inf``.
    """
    if isinstance(orders, float):
        rdp = _compute_rdp(q, noise_multiplier, orders)
    else:
        rdp = np.array([_compute_rdp(q, noise_multiplier, order) for order in orders])

    return rdp * steps


# Based on
#   https://github.com/tensorflow/privacy/blob/5f07198b66b3617b22609db983926e3ba97cd905/tensorflow_privacy/privacy/analysis/rdp_accountant.py#L237
def get_privacy_spent(orders, rdp, delta):
    """Compute epsilon given a list of RDP values and target delta.
    Args:
        orders: An array (or a scalar) of orders.
        rdp: A list (or a scalar) of RDP guarantees.
        delta: The target delta.
    Returns:
        Pair of (eps, optimal_order).
    Raises:
        ValueError: If input is malformed.
    """
    orders_vec = np.atleast_1d(orders)
    rdp_vec = np.atleast_1d(rdp)

    if delta <= 0:
        raise ValueError("Privacy failure probability bound delta must be >0.")
    if len(orders_vec) != len(rdp_vec):
        raise ValueError("Input lists must have the same length.")

    # Basic bound (see https://arxiv.org/abs/1702.07476 Proposition 3 in v3):
    #   eps = min( rdp_vec - math.log(delta) / (orders_vec - 1) )

    # Improved bound from https://arxiv.org/abs/2004.00010 Proposition 12 (in v4).
    # Also appears in https://arxiv.org/abs/2001.05990 Equation 20 (in v1).
    eps_vec = []
    for (a, r) in zip(orders_vec, rdp_vec):
        if a < 1:
            raise ValueError("Renyi divergence order must be >=1.")
        if r < 0:
            raise ValueError("Renyi divergence must be >=0.")

        if delta ** 2 + math.expm1(-r) >= 0:
            # In this case, we can simply bound via KL divergence:
            # delta <= sqrt(1-exp(-KL)).
            eps = 0  # No need to try further computation if we have eps = 0.
        elif a > 1.01:
            # This bound is not numerically stable as alpha->1.
            # Thus we have a min value of alpha.
            # The bound is also not useful for small alpha, so doesn't matter.
            eps = r + math.log1p(-1 / a) - math.log(delta * a) / (a - 1)
        else:
            # In this case we can't do anything. E.g., asking for delta = 0.
            eps = np.inf
        eps_vec.append(eps)

    idx_opt = np.argmin(eps_vec)
    return max(0, eps_vec[idx_opt]), orders_vec[idx_opt]



from typing import List

import torch
from torch.utils.data import Sampler

class DistributedPoissonSampler(Sampler):
    '''
    Distributed Poisson sampler. This is a distributed version of the Poisson sampler which selects a variable number of samples per batch (instead of a fixed number of samples per batch).
    Each batch is constructed by selecting a sample with prob sample_rate.
    The Distributed version works by assigning a subset of the dataset to each replica, and then selecting a batch with Poisson subsampling.
    So, each gpu can have a different number of samples in the batch, but the number of batches is the same.
    '''
    def __init__(self, len_dataset: int, sample_rate: float, rank: int, shuffle: bool = True, shuffle_seed: int = 42):

        self.len_dataset = len_dataset
        self.sample_rate = sample_rate
        self.rank = rank
        self.epoch = 0
        self.shuffle = shuffle
        self.shuffle_seed = shuffle_seed
        self.world_size = torch.distributed.get_world_size()
        self.generator = torch.Generator()

        self.num_samples_per_device = self.len_dataset // self.world_size
        if self.rank < self.len_dataset % self.world_size:
            # The first replicas get an extra datapoint if necessary (balanced)
            self.num_samples_per_device += 1

        # Number of batches: same as non-distributed Poisson sampling, but each batch is smaller
        self.num_batches = int(1 / self.sample_rate)

    def __iter__(self):
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.shuffle_seed + self.epoch)
            indices = torch.randperm(self.len_dataset, generator=g)
        else:
            indices = torch.arange(self.len_dataset)

        # Subset of the dataset assigned to this replica
        # (Different from the regular distributed loader that pads with more samples)
        indices = indices[self.rank : self.len_dataset : self.world_size]
        assert len(indices) == self.num_samples_per_device, f"len(indices) = {len(indices)} != {self.num_samples_per_device} = num_samples_per_device"

        # Now, select a batch with Poisson subsampling
        for _ in range(self.num_batches):
            mask = (
                torch.rand(self.num_samples_per_device, generator=self.generator)
                < self.sample_rate
            )
            selected_examples = mask.nonzero(as_tuple=False).reshape(-1)
            if len(selected_examples) > 0:
                yield indices[selected_examples]

    def __len__(self) -> int:
        """
        Expected number of batches.
        """
        return self.num_batches

    def set_epoch(self, epoch: int) -> None:
        """
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch. Otherwise, the next iteration of this
        sampler will yield the same ordering.
        Args:
            epoch (int): Epoch number.
        """
        self.epoch = epoch


# Copyright (c) Xuechen Li. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,  software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,  either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import math
from typing import Dict, Optional, Union

from . import rdp_accounting

DEFAULT_ALPHAS = tuple(1 + x / 10.0 for x in range(1, 100)) + tuple(range(12, 64))  # RDP.


class AccountingManager(abc.ABC):
    def _get_sigma_with_target_epsilon(
        self,
        target_epsilon,
        target_delta,
        sample_rate,
        steps,
        threshold,
        sigma_hi_init,
        sigma_lo_init,
    ):
        """Binary search σ given ε and δ."""
        if sigma_lo_init > sigma_hi_init:
            raise ValueError("`sigma_lo` should be smaller than `sigma_hi`.")

        # Find an appropriate region for binary search.
        sigma_hi = sigma_hi_init
        sigma_lo = sigma_lo_init

        # Ensure sigma_hi isn't too small.
        while True:
            eps = self._compute_epsilon_from_sigma(sigma_hi, sample_rate, target_delta, steps)
            if eps < target_epsilon:
                break
            sigma_hi *= 2

        # Ensure sigma_lo isn't too large.
        while True:
            eps = self._compute_epsilon_from_sigma(sigma_lo, sample_rate, target_delta, steps)
            if eps > target_epsilon:
                break
            sigma_lo /= 2

        # Binary search.
        while sigma_hi - sigma_lo > threshold:
            sigma = (sigma_hi + sigma_lo) / 2
            eps = self._compute_epsilon_from_sigma(sigma, sample_rate, target_delta, steps)
            if eps < target_epsilon:
                sigma_hi = sigma
            else:
                sigma_lo = sigma

        # Conservative estimate.
        return sigma_hi

    @abc.abstractmethod
    def compute_epsilon(self, sigma, sample_rate, target_delta, steps) -> Dict:
        """Override for reporting results."""
        raise NotImplementedError

    @abc.abstractmethod
    def _compute_epsilon_from_sigma(self, sigma, sample_rate, target_delta, steps) -> float:
        """Override for binary sigma search."""
        raise NotImplementedError

    def compute_sigma(
        self,
        target_epsilon: float,
        target_delta: float,
        sample_rate: float,
        epochs: Optional[Union[float, int]] = None,
        steps=None,
        threshold=1e-3,
        sigma_hi_init=4,
        sigma_lo_init=0.1,
    ) -> float:
        if steps is None:
            if epochs is None:
                raise ValueError("Epochs and steps cannot both be None.")
            steps = math.ceil(epochs / sample_rate)
        return self._get_sigma_with_target_epsilon(
            target_epsilon=target_epsilon,
            target_delta=target_delta,
            sample_rate=sample_rate,
            steps=steps,
            threshold=threshold,
            sigma_hi_init=sigma_hi_init,
            sigma_lo_init=sigma_lo_init,
        )


class RDPManager(AccountingManager):
    def __init__(self, alphas):
        super(RDPManager, self).__init__()
        self._alphas = alphas

    def _compute_epsilon_from_sigma(self, sigma, sample_rate, target_delta, steps):
        return self.compute_epsilon(sigma, sample_rate, target_delta, steps)["eps_rdp"]

    def compute_epsilon(self, sigma, sample_rate, target_delta, steps) -> Dict:
        """Compute RDP as usual, but convert to (ε, δ)-DP based on the result by Canonne, Kamath, Steinke."""
        rdp = rdp_accounting.compute_rdp(q=sample_rate, noise_multiplier=sigma, steps=steps, orders=self._alphas)
        eps, alpha = rdp_accounting.get_privacy_spent(orders=self._alphas, rdp=rdp, delta=target_delta)
        return dict(eps_rdp=eps, alpha_rdp=alpha)


class GLWManager(AccountingManager):
    def __init__(self, eps_error=0.05):
        super(GLWManager, self).__init__()
        self._eps_error = eps_error

    def _compute_epsilon_from_sigma(self, sigma, sample_rate, target_delta, steps):
        return self.compute_epsilon(sigma, sample_rate, target_delta, steps)["eps_upper"]  # Be conservative.

    def compute_epsilon(self, sigma, sample_rate, target_delta, steps) -> Dict:
        if steps == 0:
            return dict(eps_low=None, eps_estimate=None, eps_upper=None)

        from prv_accountant import Accountant
        accountant = Accountant(
            noise_multiplier=sigma,
            sampling_probability=sample_rate,
            delta=target_delta,
            eps_error=self._eps_error,
            max_compositions=steps
        )
        eps_low, eps_estimate, eps_upper = accountant.compute_epsilon(num_compositions=steps)
        return dict(eps_low=eps_low, eps_estimate=eps_estimate, eps_upper=eps_upper)



from . import accounting_manager
import torch

class PrivacyEngine:
    '''
    This class is used to compute the amount of noise to add to the gradients in order to satisfy the privacy budget.
    For now we only support the rdp accountant.
    '''
    def __init__(self, len_dataset: int, batch_size: int, max_grad_norm: float, num_epochs: int, target_epsilon: float, target_delta: float):
        '''
        Parameters:
            len_dataset: The number of samples in the dataset.
            batch_size: The batch size used for training.
            max_grad_norm: The maximum norm of the gradients.
            num_epochs: The number of epochs to train for.
            target_epsilon: The target epsilon value.
            target_delta: The target delta value.
        '''
        self.len_dataset = len_dataset
        self.batch_size = batch_size
        self.max_grad_norm = max_grad_norm
        self.num_epochs = num_epochs
        self.target_epsilon = target_epsilon
        self.target_delta = target_delta
        self.steps_per_epoch = len_dataset // batch_size
        self.total_steps = self.steps_per_epoch * num_epochs

        self.sample_rate = self.batch_size / self.len_dataset

        # Needed for rdp accountant
        self.alphas = accounting_manager.DEFAULT_ALPHAS

        self.manager = accounting_manager.RDPManager(alphas=self.alphas)
        self.noise_multiplier = self.manager.compute_sigma(
            target_epsilon=target_epsilon,
            target_delta=target_delta,
            sample_rate=self.sample_rate,
            epochs=self.num_epochs,
        )
        print(f"PrivacyEngine: noise_multiplier = {self.noise_multiplier}")

    def add_noise(self, grad):
        '''
        Injects noise to the gradient.
        Parameters:
            grad: The gradient of a single sample.
        Returns:
            The gradient with noise added.
        '''
        noise = torch.normal(
            mean=0.0,
            std=self.noise_multiplier * self.max_grad_norm,
            size=grad.shape,
            device=grad.device,
            dtype=grad.dtype,
        )
        return grad + noise

    def clip_gradient(self, grad):
        '''
        Clips the gradient to the maximum norm. Modify the gradient in place.
        Parameters:
            grad: The gradient of a single sample.
        '''
        torch.nn.utils.clip_grad_norm_(grad, self.max_grad_norm)